@@ -1,3 +1,10 @@
+(* TEST
+flags = " -w a "
+* setup-ocamlc.byte-build-env
+** ocamlc.byte
+*** check-ocamlc.byte-output
+*)
+
 (**************************************************************************)
 (*                                                                        *)
 (*  Crude slicer for preprocessing reachability verification tasks        *)
